/*
 *
 * LCD44780.h
 *
 *  Created on: Aug 21, 2012
 *      Author: jgnoss
 *      that's just a make *.h and *.cpp from the original only *.h
 */
// Time-stamp: "23.06.11 06:18 LCD44780.h klaus?wachtler.de"
// $Id: LCD44780.h 620 2011-07-01 05:04:42Z klaus $
//
// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
//
// [1] http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial#LCD-Ansteuerung
//
// [2] http://www.mikrocontroller.net/topic/88543#751982
//
// [3] "DOTMATRIXDISPLAYS 4x40" von
//     Electronic Assembly GmbH
//     Lochhamer Schlag 17
//     D-82166 Gräfelfing
//     http://www.lcd-module.de
//
// [4] "BLUE LINE"
//     "BLAU NEGATIV MIT WEISSER LED"
//     "DOTMATRIX LCD-Serie 1x16..4x40"
//     Manual beigelegt zu blauen LCDs von Electronic Assembly GmbH
//     Lochhamer Schlag 17
//     D-82166 Gräfelfing
//     http://www.lcd-module.de
// [5] TC1602A-08
//     http://www.abcd8.com/MANNUL/YC162A.pdf
//
// Gegenueber dem Original von [1] sind folgende wesentlichen
// Aenderungen vorgenommen worden:
// - Es kann zwanglos mehr als ein LCD gleichzeitig benutzt werden. Dabei
//   koennen alle ausser der/den Enable-Leitung(en) gemeinsam benutzt
//   werden.
// - Die Signale koennen frei ueber alle Ports verteilt werden.
// - Bei der Initialisierung muss der LCD-Typ angegeben werden.
//   Als Lohn der Muehe sollte die Positionierung auf allen Typen
//   funktionieren, wobei aber nicht alle getestet sind.
// - Der aufrufende Quelltext muss in C++ geschrieben sein, weil sonst
//   keine Referenzen und Defaultparameter moeglich waeren; zudem
//   werden die Methoden dadurch inline.
// - Sowohl Spalten als auch Zeilen werden ab 0 gezaehlt
// - Getestet mit 4x40 mit 2 Controllern
// - printf()-aehnliche Funktionen
// - selbstdefinierte Zeichen moeglich
// - ..._P-Funktionen zum Lesen aus Programmspeicher
//
//
// Bei der Auswahl des Typs (LCD_Type_...) muß man etwas aufpassen,
// wenn man
// - ein einzeiliges mit 16 Zeichen hat, da gibt es zwei Arten:
//   a) MUX 1:8, oder
//   b) MUX 1:16 (8+8); das verhält sich wie ein 2x8, bei dem die
//      beiden Zeilen hintereinander gehängt sind.
//   Dementsprechend ist LCD_Type_1x16_MUX8 oder LCD_Type_1x16_MUX16
//   auszuwählen.
// - ein 4x20 hat. Dann kann es sich einen der Controller HD44780 oder
//   K0073 oder HD66712 handeln.
//   Bei denen unterscheidet sich bei 4x20 die Adressierung (TODO,
//   siehe unten)
//
//
// Beispielverwendung: siehe AVR_LCD44780.cpp.
//
//
// Aenderungen:
//
// 23.06.2009 kw    - von aussen nicht aufzurufende Funktionen static
//                    gemacht:
//                    lcd_data(), lcd_command(), lcd_enable()
//                  - gemeinsamen Code aus lcd_command() und lcd_data()
//                    in neue Funktion lcd_byte() ausgelagert
//                  - Shiften und Maskieren in lcd_byte() eliminiert
//                  - in lcd_set_cursor() Adressberechnung korrigiert
//                    fuer blaues LCD "Blue Line" von Electronic Assembly
//                    EA W204B-NLW bzw. Reichelt LCD204BBL
//
// 02.08.2009 kw    - auf C++ umgestellt
//                  - Ports und Pinnummern beim Initialisieren
//                    anzugeben (anstatt fest verdrahtet)
//                  - mehrere LCDs nutzbar (nur Enable muss auf
//                    unterschiedlichen Pins liegen; fuer ein Display
//                    braucht man also 4 Datenpins, RS und ein oder
//                    zwei Enable; fuer jedes weitere LCD
//                    ein oder zwei weitere Enable).
//                  - Zeilen zaehlen jetzt ebenso wie Spalten ab 0
//
// 26.08.2009 kw    - Parameter fuer LCD-Typ eingefuehrt
//                  - 4x40 eingebaut und getestet
//
// 28.08.2009 kw    - selbst definierte Zeichen mit defChar()
//
// 29.08.2009 kw    - es werden intern Zeilen- und Spaltennummern
//                    mitgezählt und am Zeilenende mit Schreiben
//                    abgebrochen.
//
// 21.05.2010 kw    LCD_Type_4x20_KS0073 und LCD_Type_4x20_HD66712 ergänzt
//
//
// Aug 21, 2012 jg	- I made this a separate LCD44780.h and LCD44780.cpp so it compiles
//						right out of the box
//
//					- on the version I grabbed was an error in the two overloading
//						printSZ_P methods they didn't take the string from Flash rather
//						than from eeprom, I fixed that by renaming the eeprom methods
//						to printSZ_E so they keep reading from eeprom and I added the
//						printSZ_P reading from flash.
//
//
//
//
// Getestet mit
// - LCD:
//   - Electronic Assembly EA W204B-NLW 4x20 (LCD_Type_4x20)
//   - 2x16 aus Pollin Addon Board zu Atmel Evaluation Board (LCD_Type_2x16)
//   - 2x20
//   - 4x40 (Displaytech 404B, LCD_Type_4x40)
// - Compiler: avr-g++ 4.3.2 (Debian Linux)
//
//
// jg Version is made and tested with eclipse and
// avr-g++ (GCC) 4.5.3
//


#ifndef LCD44780_H_
#define LCD44780_H_

#include <inttypes.h>

class LCD44780 {
public:

	enum e_LCD_State {
		LCD_State_OK,
		LCD_State_TypeInvalid,
		LCD_State_PinInvalid,
		LCD_State_ColumnNumberInvalid,
		LCD_State_LineNumberInvalid,
		LCD_State_WidthTooSmall,
		LCD_State_Unknown
	};

	enum e_LCD_Type {
		LCD_Type_1x8,
		LCD_Type_1x16_MUX8,
		LCD_Type_1x16_MUX16,
		LCD_Type_1x20,
		LCD_Type_1x40,
		LCD_Type_2x8,
		LCD_Type_2x12,
		LCD_Type_2x16,
		LCD_Type_2x20,
		LCD_Type_2x24,
		LCD_Type_2x40,
		LCD_Type_4x16,
		// HD44780:
		LCD_Type_4x20,
		LCD_Type_4x20_KS0073,
		// 4x20 mit Controller HD66712, z.B. EA P204-3N/-3NLED
		// evtl. Batron BT42003 "Micro-Line"
		LCD_Type_4x20_HD66712,
		LCD_Type_4x40,
		LCD_Type_invalid
	};

	LCD44780(e_LCD_Type LCD_Type, volatile uint8_t &port_d4, uint8_t pin_d4,
			volatile uint8_t &port_d5, uint8_t pin_d5,
			volatile uint8_t &port_d6, uint8_t pin_d6,
			volatile uint8_t &port_d7, uint8_t pin_d7,
			volatile uint8_t &port_rs, uint8_t pin_rs,
			volatile uint8_t &port_e1, uint8_t pin_e1,
			volatile uint8_t &port_e2 = *(uint8_t*) 0, uint8_t pin_e2 =
					(uint8_t) -1);
	virtual ~LCD44780();

	// Loescht letzte Fehlernummer:
	void resetState();

	// liefert den Status bzw. den letzten Fehler
	e_LCD_State getState();

	// liefert aktuelle Zeilennummer
	uint8_t getLineNumber();

	// liefert aktuelle Spaltennummer
	uint8_t getColumnNumber();

	// Loescht das Display
	void clear();

	// Sendet den Befehl: Cursor Home
	void home();

	// setzt den Cursor in Zeile y (0..3) Spalte x (0..39)
	void setCursor(uint8_t x, uint8_t y);

	// Schreibt einen nullterminierten String auf das LCD
	void printSZ(const char *p_data);

	// Schreibt einen String aus dem EEprom auf das LCD
	void printSZ_E(const char *p_data);

	// Schreibt einen String aus dem Flash auf das LCD
	void printSZ_P(const char *p_data);

	// Schreibt einen String an der Stelle (x,y) auf das LCD
	void printSZ(uint8_t x, uint8_t y, const char *p_data);

	// Schreibt einen String aus dem EEprom auf das LCD
	void printSZ_E(uint8_t x, uint8_t y, const char *p_data);

	// Schreibt einen String aus dem Flash auf das LCD
	void printSZ_P(uint8_t x, uint8_t y, const char *p_data);

	// sendet ein Datenbyte an das LCD
	void printChar(unsigned char data);

	// setzt Cursor und sendet ein Datenbyte an das LCD
	void printChar(uint8_t x, uint8_t y, unsigned char data);

	// definiert ein Zeichen im Zeichensatz an der Position pos (0 bis
	// 7) mit den Bitdaten aus arrBitmuster[].
	// arrBitmuster[] enthaelt 8 uint8_t, wobei das erste Byte in
	// seinen unteren 4 Bit die oberste Zeile des 5x8 Pixel grossen
	// Zeichens enthaelt usw. bis arrBitmuster[7], das in seinen
	// unteren 5 Bit die unterste Pixelzeile enthaelt.
	// Bei einem 2x16 aus dem Pollin Addon-Board kann ich nur die
	// ersten 4 Zeichen definieren; bei allen anderen waren es bisher
	// immer 8.
	void defChar(uint8_t pos, uint8_t arrBitmuster[8]);

	// gibt an der aktuellen Cursorposition aus, sonst wie
	// std::printf().
	// Rueckgabe: Anzahl der ausgegebenen Zeichen (0, wenn es mehr als
	// geworden worden waeren, als in die Zeile passt; in diesem Fall
	// werden '#' ausgegeben)
	uint8_t printf(const char *fmt, ...);

	// gibt an der aktuellen Cursorposition aus, sonst wie
	// std::printf().
	// fmt ist ein Zeiger in den Programmspeicher.
	// Rueckgabe: Anzahl der ausgegebenen Zeichen (0, wenn es mehr als 40
	// geworden worde waeren, als in die Zeile passt; in diesem Fall
	// werden '#' ausgegeben)
	uint8_t printf_P(const char *fmt, ...);

	// gibt an der Cursorposition (x,y) n Zeichen aus, sonst wie std::printf().
	// Wenn die Ausgabe gem. fmt nicht zu n Zeichen fuehrt, wird rechts mit
	// Leerzeichen aufgefuellt.
	// Rueckgabe: Anzahl der ausgegebenen Zeichen ohne die aufgefuellten
	// Leerzeichen (0, wenn es mehr als n geworden waeren oder wenn
	// n>Zeilenlaenge; in diesen Faellen werden '#' ausgegeben)
	uint8_t printf(uint8_t x, uint8_t y, uint8_t n, const char *fmt, ...);

	// gibt an der Cursorposition (x,y) n Zeichen aus, sonst wie std::printf().
	// Wenn die Ausgabe gem. fmt nicht zu n Zeichen fuehrt, wird rechts mit
	// Leerzeichen aufgefuellt.
	// fmt ist ein Zeiger in den Programmspeicher.
	// Rueckgabe: Anzahl der ausgegebenen Zeichen ohne die aufgefuellten
	// Leerzeichen (0, wenn es mehr als n geworden waeren oder wenn
	// n>Zeilenlaenge; in diesen Faellen werden '#' ausgegeben)
	int printf_P(uint8_t x, uint8_t y, uint8_t n, const char *fmt, ...);

private:

	// waehlt den ersten (0) oder zweiten (1) Controller aus
	void selectController(uint8_t iCtrlr);

	// wechselt den Controller (1 , wenn vorher 0; 0 wenn vorher 1)
	void changeController();

	// erzeugt den Enable-Puls fuer den Controller, der zuletzt mit
	// selectController() ausgewaehlt wurde
	void enable();

	// sendet ein Daten- oder Kommandobyte (RS am LCD muss vorher
	// entsprechend gesetzt sein)
	void writeByte(unsigned char data);

	// sendet einen Befehl an das LCD
	void writeCommand(unsigned char command);

	enum {
		LCD_CLEAR_DISPLAY = 0x01, LCD_CURSOR_HOME = 0x02,
	};

	// aktueller Zustand (LCD_State_OK oder ein Fehler):
	e_LCD_State LCD_State;

	// Typ (4x20 etc.)
	e_LCD_Type LCD_Type;

	// Anzahl Zeilen, Spalten:
	uint8_t nZeilen;
	uint8_t nSpalten;

	// aktuelle Zeile, Spalte:
	uint8_t iZeile;
	uint8_t iSpalte;

	// Ansteuerung (die port_... sind Referenzen auf PORTx; die
	// pin_... sind Pinnummern 0...7):
	volatile uint8_t &port_e1;
	uint8_t pin_e1;
	volatile uint8_t &port_e2;
	uint8_t pin_e2;
	volatile uint8_t &port_rs;
	uint8_t pin_rs;
	volatile uint8_t &port_d4;
	uint8_t pin_d4;
	volatile uint8_t &port_d5;
	uint8_t pin_d5;
	volatile uint8_t &port_d6;
	uint8_t pin_d6;
	volatile uint8_t &port_d7;
	uint8_t pin_d7;

	// jeweils aktuelle Zeiger auf port_e1 und pin_e1 oder port_e2 und pin_e2:
	volatile uint8_t *p_port_e;
	uint8_t *p_pin_e;

	// Tabelle, welche Zeile an welcher Stelle im DD-RAM beginnt (wird
	// je nach LCD_Type initalisiert):
	uint8_t map_Linestart_DDRAM[4];

};

#endif /* LCD44780_H_ */

